# -*- coding: utf-8 -*-
from __future__ import print_function
import mimetypes
mimetypes.init()

NEW_MIMETYPES = {
    '.dwg': 'image/x-dwg',
    '.ico': 'image/x-icon',
    '.bz2': 'application/x-bzip2',
    '.gz': 'application/x-gzip',
    '.xlsx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.sheet',
    '.xltx': 'application/vnd.openxmlformats-officedocument.spreadsheetml.template',
    '.potx': 'application/vnd.openxmlformats-officedocument.presentationml.template',
    '.ppsx': 'application/vnd.openxmlformats-officedocument.presentationml.slideshow',
    '.pptx': 'application/vnd.openxmlformats-officedocument.presentationml.presentation',
    '.sldx': 'application/vnd.openxmlformats-officedocument.presentationml.slide',
    '.docx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.document',
    '.dotx': 'application/vnd.openxmlformats-officedocument.wordprocessingml.template',
    '.xlam': 'application/vnd.ms-excel.addin.macroEnabled.12',
    '.xlsb': 'application/vnd.ms-excel.sheet.binary.macroEnabled.12',
}
for name, value in NEW_MIMETYPES.items():
    mimetypes.types_map[name] = value

import os
from zlib import adler32

from cherrypy.lib import httputil, file_generator_limited
from werkzeug.http import quote_etag
from werkzeug.exceptions import RequestedRangeNotSatisfiable

from .helpers import to_unicode


def serve_file(path):
    headers = {}

    st = os.stat(path)

    etag = 'clay-{0}-{1}-{2}'.format(
        os.path.getmtime(path),
        os.path.getsize(path),
        adler32(path.encode('utf-8')) & 0xffffffff
    )
    headers['ETag'] = quote_etag(etag)

    # Set the Last-Modified response header, so that
    # modified-since validation code can work.
    headers['Last-Modified'] = httputil.HTTPDate(st.st_mtime)

    _, ext = os.path.splitext(path)
    content_type = mimetypes.types_map.get(ext, None)
    headers['Content-Type'] = content_type or 'text/plain'

    fileobj = open(path, 'rb')
    return serve_fileobj(fileobj, headers, st.st_size)


def serve_fileobj(fileobj, headers, content_length):
    status_code = 200
    headers["Accept-Ranges"] = "bytes"

    r = httputil.get_ranges(headers.get('Range'), content_length)

    if r == []:
        headers['Content-Range'] = "bytes */{0}".format(content_length)
        message = "Invalid Range (first-byte-pos greater than Content-Length)"
        raise RequestedRangeNotSatisfiable(message)

    if not r:
        headers['Content-Length'] = content_length
        return fileobj, headers, status_code

    # Return a multipart/byteranges response.
    status_code = 206

    if len(r) == 1:
        # Return a single-part response.
        start, stop = r[0]
        if stop > content_length:
            stop = content_length
        r_len = stop - start

        headers['Content-Range'] = "bytes {0}-{1}/{2}".format(
            start, stop - 1, content_length
        )
        headers['Content-Length'] = r_len
        fileobj.seek(start)
        body = file_generator_limited(fileobj, r_len)
        return body, headers, status_code

    try:
        # Python 3
        from email.generator import _make_boundary as make_boundary
    except ImportError:
        # Python 2
        from mimetools import choose_boundary as make_boundary

    boundary = make_boundary()
    content_type = "multipart/byteranges; boundary={0}".format(boundary)
    headers['Content-Type'] = content_type
    if "Content-Length" in headers:
        del headers["Content-Length"]

    def file_ranges():
        for start, stop in r:
            yield to_unicode("--" + boundary)
            yield to_unicode("\r\nContent-type: {0}".format(content_type))
            yield to_unicode(
                "\r\nContent-range: bytes {0}-{1}/{2}\r\n\r\n".format(
                    start, stop - 1, content_length
                )
            )
            fileobj.seek(start)

            gen = file_generator_limited(fileobj, stop - start)
            for chunk in gen:
                yield chunk
            yield to_unicode("\r\n")

        yield to_unicode("--" + boundary + "--")

    body = file_ranges()
    return body, headers, status_code



# -*- coding: utf-8 -*-
from collections import OrderedDict
from fnmatch import fnmatch
from functools import partial
import os
from os.path import dirname, abspath, isdir, isfile, join
import re

from flask import request

from .helpers import to_unicode


def norm_url(url):
    url = url.strip().rstrip('/')
    url = re.sub('index.html$', '', url).rstrip('/')
    if url.startswith('/'):
        return url
    baseurl = dirname(request.path.strip('/'))
    if baseurl:
        return '/' + '/'.join([baseurl, url])
    return '/' + url


def active(*url_patterns, **kwargs):
    partial = kwargs.get('partial')

    path = norm_url(request.path)

    # Accept single patterns also
    if len(url_patterns) == 1 and isinstance(url_patterns[0], (list, tuple)):
        url_patterns = url_patterns[0]

    for urlp in url_patterns:
        urlp = norm_url(urlp)
        if fnmatch(path, urlp) or (partial and path.startswith(urlp)):
            return 'active'
    return u''


class ToC(object):
    """Recursevly builds a tree of (name, relpath) of all files and subfolders
    of a given path.

    Usage:

        clay
        ├── __init__.py
        ├── helpers.py
        ├── main.py
        └── source
            ├── _index.html
            ├── _index.txt
            └── foo
                └── bar.html


        >>> toc = TOC('/Users/bla/Projects/clay/clay')
        >>> list(toc)
        [
            (u'__init__.py', u'/__init__.py'),
            (u'helpers.py', u'/helpers.py'),
            (u'main.py', u'/main.py'),
        ]
        >>> list(toc.source)
        [
            (u'_index.html', u'/source/_index.html'),
            (u'_index.txt', u'/source/_index.txt'),
            (u'_notfound.html', u'/source/_notfound.html'),
        ]
        >>> toc(maxdepth=1)
        '''<ul class="toc">
            <li><a href="/__init__.py">__init__.py</a></li>
            <li><a href="/helpers.py">helpers.py</a></li>
            <li><a href="/main.py">main.py</a></li>
            <li><span>source</span>
                <ul>
                    <li><a href="/source/_index.html">_index</a></li>
                    <li><a href="/source/_index.txt">_index.txt</a></li>
                </ul>
            </li>
        </ul>'''
        >>> toc.source(maxdepth=1)
        '''<ul class="toc">
            <li><a href="/source/_index.html">_index</a></li>
            <li><a href="/source/_index.txt">_index.txt</a></li>
            <li><span>foo</span>
                <ul>
                    <li><a href="/source/foo/bar.html">bar.html</a></li>
                </ul>
            </li>
        </ul>'''
        >>> toc(maxdepth=1, folders_first=True)
        '''<ul class="toc">
            <li><span>source</span>
                <ul class="toc">
                    <li><a href="/source/_index.html">_index</a></li>
                    <li><a href="/source/_index.txt">_index.txt</a></li>
                </ul>
            </li>
            <li><a href="/__init__.py">__init__.py</a></li>
            <li><a href="/helpers.py">helpers.py</a></li>
            <li><a href="/main.py">main.py</a></li>
        </ul>'''

    """
    def __init__(self, basepath, baseurl='/', filter=('.*', '*.pyc')):
        self._leafs = []
        self._branches = []
        basepath = abspath(to_unicode(basepath))
        if isdir(basepath):
            self.__buildtree(basepath, baseurl, filter)

    def __buildtree(self, basepath, baseurl, filter):
        leafs = OrderedDict()
        branches = OrderedDict()

        names = sorted(os.listdir(basepath))
        for name in names:
            if any(map(partial(fnmatch, name), filter)):
                continue
            if isfile(join(basepath, name)):
                leafs[name] = join(baseurl, name)
            else:
                new_basepath = join(basepath, name)
                new_baseurl = join(baseurl, name)
                branches[name] = ToC(new_basepath, baseurl=new_baseurl)

        self._leafs = leafs
        self._branches = branches

    def __getattr__(self, name):
        node = self._leafs.get(name)
        if not node:
            node = self._branches.get(name)
            if not node:
                raise AttributeError
        return node

    def __getitem__(self, key):
        return self.__getattr__(key)

    def __iter__(self):
        return self._leafs.iteritems()

    def __call__(self, maxdepth=1, folders_first=False, indent=4, _depth=0):
        html_leafs = self._render_leafs(
            indent=indent,
            _depth=_depth
        )
        html_branches = u''
        if maxdepth:
            html_branches = self._render_branches(
                maxdepth=(maxdepth - 1),
                folders_first=folders_first,
                indent=indent,
                _depth=_depth
            )
        tmpl = u'{indent}<ul class="toc">\n{html1}{html2}{indent}</ul>'
        if folders_first:
            html1 = html_branches
            html2 = html_leafs
        else:
            html1 = html_leafs
            html2 = html_branches
        return tmpl.format(
            html1=html1,
            html2=html2,
            indent=u' ' * (indent * _depth)
        )

    def _render_leafs(self, indent, _depth):
        lis = []
        _indent = u' ' * (indent * (_depth + 1))

        for name, path in self._leafs.items():
            html = u'{indent}<li><a href="{path}">{name}</a></li>'.format(
                indent=_indent, name=name, path=path
            )
            lis.append(html)

        if not lis:
            return u''
        return u'\n'.join(lis) + u'\n'

    def _render_branches(self, maxdepth, folders_first, indent, _depth):
        uls = []
        for name, toc in self._branches.items():
            tmpl = u'{indent}<li><span>{name}</span>\n{subtoc}\n{indent}</li>'
            ulhtml = tmpl.format(
                name=name,
                indent=u' ' * (indent * (_depth + 1)),
                subtoc=toc(
                    maxdepth=maxdepth - 1,
                    folders_first=folders_first,
                    indent=indent,
                    _depth=_depth + 2
                )
            )
            uls.append(ulhtml)

        if not uls:
            return u''
        return u'\n'.join(uls) + u'\n'



# -*- coding: utf-8 -*-
"""
# Clay

An amazing prototyping tool

------------
© [Juan-Pablo Scaletti](http://jpscaletti.com).
[MIT License](http://www.opensource.org/licenses/mit-license.php)

"""
from .main import Clay  # noqa


__version__ = '2.12.4'



# -*- coding: utf-8 -*-
from __future__ import print_function

from os.path import dirname, join

from clay.tglobals import ToC


CLAY_SOURCE_PATH = join(dirname(__file__), u'..', u'clay')
FILTER = ('.*', '*.pyc', )


def test_toc_build():
    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)

    assert toc._leafs.items() == [
        (u'__init__.py', u'/__init__.py'),
        (u'helpers.py', u'/helpers.py'),
        (u'jinja_includewith.py', u'/jinja_includewith.py'),
        (u'main.py', u'/main.py'),
        (u'manage.py', u'/manage.py'),
        (u'server.py', u'/server.py'),
        (u'static.py', u'/static.py'),
        (u'tglobals.py', u'/tglobals.py'),
        (u'wsgiapp.py', u'/wsgiapp.py')
    ]
    assert toc._branches.keys() == [u'markdown_ext', u'source']


def test_toc_iter():
    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)

    assert list(toc) == [
        (u'__init__.py', u'/__init__.py'),
        (u'helpers.py', u'/helpers.py'),
        (u'jinja_includewith.py', u'/jinja_includewith.py'),
        (u'main.py', u'/main.py'),
        (u'manage.py', u'/manage.py'),
        (u'server.py', u'/server.py'),
        (u'static.py', u'/static.py'),
        (u'tglobals.py', u'/tglobals.py'),
        (u'wsgiapp.py', u'/wsgiapp.py')
    ]
    assert list(toc.source) == [
        (u'_index.html', u'/source/_index.html'),
        (u'_index.txt', u'/source/_index.txt'),
        (u'_notfound.html', u'/source/_notfound.html'),
    ]


def test_toc_render_maxdepth0():
    expected = """<ul class="toc">
    <li><a href="/__init__.py">__init__.py</a></li>
    <li><a href="/helpers.py">helpers.py</a></li>
    <li><a href="/jinja_includewith.py">jinja_includewith.py</a></li>
    <li><a href="/main.py">main.py</a></li>
    <li><a href="/manage.py">manage.py</a></li>
    <li><a href="/server.py">server.py</a></li>
    <li><a href="/static.py">static.py</a></li>
    <li><a href="/tglobals.py">tglobals.py</a></li>
    <li><a href="/wsgiapp.py">wsgiapp.py</a></li>
</ul>"""

    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)
    html = toc(maxdepth=0)
    assert expected == html


def test_toc_render_maxdepth1():
    expected = """<ul class="toc">
    <li><a href="/__init__.py">__init__.py</a></li>
    <li><a href="/helpers.py">helpers.py</a></li>
    <li><a href="/jinja_includewith.py">jinja_includewith.py</a></li>
    <li><a href="/main.py">main.py</a></li>
    <li><a href="/manage.py">manage.py</a></li>
    <li><a href="/server.py">server.py</a></li>
    <li><a href="/static.py">static.py</a></li>
    <li><a href="/tglobals.py">tglobals.py</a></li>
    <li><a href="/wsgiapp.py">wsgiapp.py</a></li>
    <li><span>markdown_ext</span>
        <ul class="toc">
            <li><a href="/markdown_ext/__init__.py">__init__.py</a></li>
            <li><a href="/markdown_ext/jinja.py">jinja.py</a></li>
            <li><a href="/markdown_ext/md_admonition.py">md_admonition.py</a></li>
            <li><a href="/markdown_ext/md_captions.py">md_captions.py</a></li>
            <li><a href="/markdown_ext/md_delinsmark.py">md_delinsmark.py</a></li>
            <li><a href="/markdown_ext/md_fencedcode.py">md_fencedcode.py</a></li>
            <li><a href="/markdown_ext/md_superscript.py">md_superscript.py</a></li>
            <li><a href="/markdown_ext/render.py">render.py</a></li>
        </ul>
    </li>
    <li><span>source</span>
        <ul class="toc">
            <li><a href="/source/_index.html">_index.html</a></li>
            <li><a href="/source/_index.txt">_index.txt</a></li>
            <li><a href="/source/_notfound.html">_notfound.html</a></li>
        </ul>
    </li>
</ul>"""

    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)
    html = toc(maxdepth=1)
    print(html)
    assert expected == html


def test_toc_render_maxdepthplus():
    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)
    assert toc(maxdepth=2) == toc(maxdepth=9999)


def test_toc_render_folders_first():
    expected = """<ul class="toc">
    <li><span>markdown_ext</span>
        <ul class="toc">
            <li><a href="/markdown_ext/__init__.py">__init__.py</a></li>
            <li><a href="/markdown_ext/jinja.py">jinja.py</a></li>
            <li><a href="/markdown_ext/md_admonition.py">md_admonition.py</a></li>
            <li><a href="/markdown_ext/md_captions.py">md_captions.py</a></li>
            <li><a href="/markdown_ext/md_delinsmark.py">md_delinsmark.py</a></li>
            <li><a href="/markdown_ext/md_fencedcode.py">md_fencedcode.py</a></li>
            <li><a href="/markdown_ext/md_superscript.py">md_superscript.py</a></li>
            <li><a href="/markdown_ext/render.py">render.py</a></li>
        </ul>
    </li>
    <li><span>source</span>
        <ul class="toc">
            <li><a href="/source/_index.html">_index.html</a></li>
            <li><a href="/source/_index.txt">_index.txt</a></li>
            <li><a href="/source/_notfound.html">_notfound.html</a></li>
        </ul>
    </li>
    <li><a href="/__init__.py">__init__.py</a></li>
    <li><a href="/helpers.py">helpers.py</a></li>
    <li><a href="/jinja_includewith.py">jinja_includewith.py</a></li>
    <li><a href="/main.py">main.py</a></li>
    <li><a href="/manage.py">manage.py</a></li>
    <li><a href="/server.py">server.py</a></li>
    <li><a href="/static.py">static.py</a></li>
    <li><a href="/tglobals.py">tglobals.py</a></li>
    <li><a href="/wsgiapp.py">wsgiapp.py</a></li>
</ul>"""

    toc = ToC(CLAY_SOURCE_PATH, filter=FILTER)
    html = toc(maxdepth=1, folders_first=True)
    print(html)
    assert expected == html
